@@ -52,6 +52,7 @@ struct GSWindowPrivate
 
         guint      lock_enabled : 1;
         guint      logout_enabled : 1;
+        guint      user_switch_enabled : 1;
         guint64    logout_timeout;
 
         GtkWidget *box;
@@ -68,6 +69,8 @@ struct GSWindowPrivate
         gint       watch_id;
         gint       dialog_response;
 
+        GList     *key_events;
+
         GTimer    *timer;
 };
 
@@ -379,11 +382,44 @@ plug_removed (GtkWidget *widget,
         return TRUE;
 }
 
+static void
+forward_key_events (GSWindow *window)
+{
+        window->priv->key_events = g_list_reverse (window->priv->key_events);
+
+        while (window->priv->key_events) {
+                GdkEventKey *event = window->priv->key_events->data;
+
+                gtk_window_propagate_key_event (GTK_WINDOW (window), event);
+
+                gdk_event_free ((GdkEvent *)event);
+                window->priv->key_events = g_list_delete_link (window->priv->key_events,
+                                                               window->priv->key_events);
+        }
+}
+
+static void
+remove_key_events (GSWindow *window)
+{
+        window->priv->key_events = g_list_reverse (window->priv->key_events);
+
+        while (window->priv->key_events) {
+                GdkEventKey *event = window->priv->key_events->data;
+
+                gdk_event_free ((GdkEvent *)event);
+                window->priv->key_events = g_list_delete_link (window->priv->key_events,
+                                                               window->priv->key_events);
+        }
+}
+
 static void
 socket_show (GtkWidget *widget,
              GSWindow  *window)
 {
         gtk_widget_child_focus (window->priv->socket, GTK_DIR_TAB_FORWARD);
+
+        /* send queued events to the dialog */
+        forward_key_events (window);
 }
 
 static void
@@ -449,6 +485,9 @@ gs_window_dialog_finish (GSWindow *window)
 
         g_spawn_close_pid (window->priv->pid);
         window->priv->pid = 0;
+
+        /* remove events for the case were we failed to show socket */
+        remove_key_events (window);
 }
 
 static gboolean
@@ -537,35 +576,43 @@ is_logout_enabled (GSWindow *window)
         return FALSE;
 }
 
+static gboolean
+is_user_switch_enabled (GSWindow *window)
+{
+  	return window->priv->user_switch_enabled;
+}
+
 static gboolean
 popup_dialog_idle (GSWindow *window)
 {
         gboolean  result;
-        char     *command;
+        char     *tmp;
+        GString  *command;
 
-        command = g_build_filename (LIBEXECDIR, "gnome-screensaver-dialog", NULL);
+        tmp = g_build_filename (LIBEXECDIR, "gnome-screensaver-dialog", NULL);
+        command = g_string_new (tmp);
+        g_free (tmp);
 
         if (is_logout_enabled (window)) {
-                char *cmd;
-
-                cmd = g_strdup_printf ("%s --enable-logout", command);
-                g_free (command);
-                command = cmd;
+                command = g_string_append (command, " --enable-logout");
+        }
+        if (is_user_switch_enabled (window)) {
+                command = g_string_append (command, " --enable-switch");
         }
 
         gs_window_clear (window);
         set_invisible_cursor (GTK_WIDGET (window)->window, FALSE);
 
         result = spawn_on_window (window,
-                                  command,
+                                  command->str,
                                   &window->priv->pid,
                                   (GIOFunc)command_watch,
                                   window,
                                   &window->priv->watch_id);
         if (!result)
-                g_warning ("Could not start command: %s", command);
+                g_warning ("Could not start command: %s", command->str);
 
-        g_free (command);
+        g_string_free (command, TRUE);
 
         window->priv->popup_dialog_idle_id = 0;
 
@@ -633,6 +680,15 @@ gs_window_set_logout_enabled (GSWindow *window,
         window->priv->logout_enabled = logout_enabled;
 }
 
+void
+gs_window_set_user_switch_enabled (GSWindow *window,
+                                   gboolean  user_switch_enabled)
+{
+        g_return_if_fail (GS_IS_WINDOW (window));
+
+        window->priv->user_switch_enabled = user_switch_enabled;
+}
+
 void
 gs_window_set_logout_timeout (GSWindow *window,
                               glong     logout_timeout)
@@ -737,14 +793,38 @@ gs_window_request_unlock_idle (GSWindow *window)
         return FALSE;
 }
 
+static void
+queue_key_event (GSWindow    *window,
+                 GdkEventKey *event)
+{
+        window->priv->key_events = g_list_prepend (window->priv->key_events,
+                                                   gdk_event_copy ((GdkEvent *)event));
+}
+
 static gboolean
 gs_window_real_key_press_event (GtkWidget   *widget,
                                 GdkEventKey *event)
 {
+        gboolean catch_events = FALSE;
+
         /*g_message ("KEY PRESS state: %u keyval %u", event->state, event->keyval);*/
 
-        if (GS_WINDOW (widget)->priv->request_unlock_idle_id == 0) {
-                GS_WINDOW (widget)->priv->request_unlock_idle_id = g_idle_add ((GSourceFunc)gs_window_request_unlock_idle, widget);
+        /* if we don't already have a socket then request an unlock */
+        if (! GS_WINDOW (widget)->priv->socket) {
+                if (GS_WINDOW (widget)->priv->request_unlock_idle_id == 0) {
+                        GS_WINDOW (widget)->priv->request_unlock_idle_id = g_idle_add ((GSourceFunc)gs_window_request_unlock_idle, widget);
+                }
+
+                catch_events = TRUE;
+        } else {
+                if (! GTK_WIDGET_VISIBLE (GS_WINDOW (widget)->priv->socket)) {
+                        catch_events = TRUE;
+                }
+        }
+
+        /* Catch all keypresses up until the lock dialog is shown */
+        if (catch_events) {
+                queue_key_event (GS_WINDOW (widget), event);
         }
 
         if (GTK_WIDGET_CLASS (parent_class)->key_press_event)
@@ -757,7 +837,10 @@ static gboolean
 gs_window_real_motion_notify_event (GtkWidget      *widget,
                                     GdkEventMotion *event)
 {
-        if (GS_WINDOW (widget)->priv->request_unlock_idle_id == 0) {
+
+        /* if we don't already have a socket then request an unlock */
+        if (! GS_WINDOW (widget)->priv->socket
+            && (GS_WINDOW (widget)->priv->request_unlock_idle_id == 0)) {
                 GS_WINDOW (widget)->priv->request_unlock_idle_id = g_idle_add ((GSourceFunc)gs_window_request_unlock_idle, widget);
         }
 
@@ -922,6 +1005,9 @@ gs_window_finalize (GObject *object)
         if (window->priv->timer)
                 g_timer_destroy (window->priv->timer);
 
+        /* just in case they weren't removed */
+        remove_key_events (window);
+
         G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
